import { ReactElement, useContext } from 'react';
import { Line } from 'react-chartjs-2';

import {
    Chart as ChartJS,
    CategoryScale,
    LinearScale,
    PointElement,
    LineElement,
    Title,
    Tooltip,
    Legend,
} from 'chart.js';

import { DarkFibreLease } from "compendium/Schema";
import { createLineChartDataset } from "compendium/helpers/dataconversion";
import DataPage from 'compendium/components/DataPage';
import Filter from "compendium/components/graphing/Filter";
import { Sections } from "compendium/helpers/constants";
import { FilterSelectionContext } from 'compendium/providers/FilterSelectionProvider';
import ChartContainer from "compendium/components/graphing/ChartContainer";
import { useData } from 'compendium/helpers/useData';
import { getLineChartOptions } from 'compendium/helpers/charthelpers';
import titles from "compendium/titles";

ChartJS.register(
    CategoryScale,
    LinearScale,
    PointElement,
    LineElement,
    Title,
    Tooltip,
    Legend
);

function IRUDurationPage(): ReactElement {

    const field = 'iru_duration';
    const validityCheck = (data: DarkFibreLease) => data[field] != null
    const { filterSelection, setFilterSelection } = useContext(FilterSelectionContext);
    const { data, nrens } = useData<DarkFibreLease>('/api/dark-fibre-lease', setFilterSelection, validityCheck);

    const selectedData = data.filter(data =>
        filterSelection.selectedNrens.includes(data.nren) // we only allow filtering nrens for this page
    );

    const iruDurationData = createLineChartDataset(selectedData, field);

    const filterNode = <Filter
        filterOptions={{ availableYears: [], availableNrens: [...nrens.values()] }}
        filterSelection={filterSelection}
        setFilterSelection={setFilterSelection}
    />

    const options = getLineChartOptions({ title: 'Lease Duration In Years', tooltipUnit: 'years', tickLimit: 999 });

    return (
        <DataPage title={titles['iru-duration']}
            description={<span>NRENs sometimes take out an IRU (Indefeasible Right of Use),
                which is essentially a long-term lease, on a portion of the capacity of a cable
                rather than laying cable themselves. This graph shows the average duration,
                in years, of the IRUs of NRENs.</span>}
            category={Sections.Network} filter={filterNode}
            data={selectedData} filename="iru_duration_data">
            <ChartContainer>
                <Line data={iruDurationData} options={options} />
            </ChartContainer>
        </DataPage>
    );
}

export default IRUDurationPage;
