import React, { useContext } from "react";
import { Standards } from "compendium/Schema";
import { createNrenTableLookup, addTooltip } from "compendium/helpers/dataconversion";
import DataPage from "compendium/components/DataPage";
import Filter from "compendium/components/graphing/Filter";
import { Sections } from "compendium/helpers/constants";
import { FilterSelectionContext } from "compendium/providers/FilterSelectionProvider";
import ChartContainer from "compendium/components/graphing/ChartContainer";
import { useData } from "compendium/helpers/useData";
import PillTable from "compendium/components/PillTable";
import titles from "compendium/titles";

function AuditsPage(): React.ReactElement {
    const dataField = 'audits';
    const validityCheck = (data: Standards) => data[dataField] !== null;
    const { filterSelection, setFilterSelection } = useContext(FilterSelectionContext);
    const { data, years, nrens } = useData<Standards>('/api/standards', setFilterSelection, validityCheck);

    const selectedData = data.filter(data =>
        filterSelection.selectedYears.includes(data.year) && filterSelection.selectedNrens.includes(data.nren) && data[dataField] !== null
    );

    const dataLookup = createNrenTableLookup(selectedData, dataField);

    const processTooltip = (column: string, standards: Standards) => {
        if (!standards.audit_specifics) {
            return;
        }
        return standards.audit_specifics;
    }

    const withTooltip = addTooltip(dataLookup, processTooltip);

    const columns = [
        "Yes",
        "No",
    ]

    const columnValueMap = new Map<string, string>([
        [columns[0], "True"],
        [columns[1], "False"],
    ]);

    const filterNode = <Filter
        filterOptions={{ availableYears: [...years], availableNrens: [...nrens.values()] }}
        filterSelection={filterSelection}
        setFilterSelection={setFilterSelection}
        coloredYears
    />

    return (
        <DataPage title={titles.audits}
            description="The table below shows whether NRENs have external and/or internal audits 
            of the information security management systems (eg. risk management and policies). 
            Where extra information has been provided, such as whether a certified security auditor 
            on ISP 27001 is performing the audits, it can be viewed by hovering over the indicator 
            mark ringed in black."
            category={Sections.Policy} filter={filterNode}
            data={selectedData} filename="audits_nrens_per_year">
            <ChartContainer>
                <PillTable columns={columns} columnLookup={columnValueMap} dataLookup={withTooltip} />
            </ChartContainer>
        </DataPage>
    );
}
export default AuditsPage;
