import React, { useContext } from "react";
import { Standards } from "compendium/Schema";
import { createNrenTableLookup, addTooltip } from "compendium/helpers/dataconversion";
import DataPage from "compendium/components/DataPage";
import Filter from "compendium/components/graphing/Filter";
import { Sections } from "compendium/helpers/constants";
import { FilterSelectionContext } from "compendium/providers/FilterSelectionProvider";
import ChartContainer from "compendium/components/graphing/ChartContainer";
import { useData } from "compendium/helpers/useData";
import PillTable from "compendium/components/PillTable";
import titles from "compendium/titles";

function BusinessContinuityPage(): React.ReactElement {
    const dataField = 'business_continuity_plans';
    const validityCheck = (data: Standards) => data[dataField] !== null;
    const { filterSelection, setFilterSelection } = useContext(FilterSelectionContext);
    const { data, years, nrens } = useData<Standards>('/api/standards', setFilterSelection, validityCheck);

    const selectedData = data.filter(data =>
        filterSelection.selectedYears.includes(data.year) && filterSelection.selectedNrens.includes(data.nren) && data[dataField] !== null
    );

    const dataLookup = createNrenTableLookup(selectedData, dataField);

    const processTooltip = (column: string, standards: Standards) => {
        if (!standards.business_continuity_plans_specifics) {
            return;
        }
        return standards.business_continuity_plans_specifics;
    }

    const withTooltip = addTooltip(dataLookup, processTooltip);

    const columns = [
        "Yes",
        "No",
    ]

    const columnValueMap = new Map<string, string>([
        [columns[0], "True"],
        [columns[1], "False"],
    ]);

    const filterNode = <Filter
        filterOptions={{ availableYears: [...years], availableNrens: [...nrens.values()] }}
        filterSelection={filterSelection}
        setFilterSelection={setFilterSelection}
        coloredYears
    />

    return (
        <DataPage title={titles["business-continuity"]}
            description="The table below shows which NRENs have business continuity plans in place to 
            ensure business continuation and operations. Extra details about whether the NREN 
            complies with any international standards, and whether they test the continuity plans 
            regularly can be seen by hovering over the marker. The presence of this extra information 
            is denoted by a black ring around the marker."
            category={Sections.Policy} filter={filterNode}
            data={selectedData} filename="business_continuity_nrens_per_year">
            <ChartContainer>
                <PillTable columns={columns} columnLookup={columnValueMap} dataLookup={withTooltip} />
            </ChartContainer>
        </DataPage>
    );
}
export default BusinessContinuityPage;
