import React, { useState, useEffect, useContext } from "react";
import { Container, Table, Row, Form, InputGroup, Accordion, Button } from "react-bootstrap";
import toast, { Toaster } from "react-hot-toast";
import { userContext } from "compendium/providers/UserProvider";
import { useConfig } from "compendium/helpers/useConfig";
import { User, Nren } from "compendium/Schema";
import SurveySidebar from "compendium/survey/management/SurveySidebar";
import { FaCheck } from "react-icons/fa";
import { debounce } from "lodash";

async function fetchUsers(): Promise<User[]> {
    try {
        const response = await fetch("/api/user/list");
        return await response.json()
    } catch {
        return [];
    }
}

async function fetchNrens(): Promise<Nren[]> {
    try {
        const response = await fetch("/api/nren/list");
        return await response.json()
    } catch {
        return [];
    }
}

async function updateUser(id, changes): Promise<User> {

    const body = {
        id: id,
        ...changes
    }
    const requestOptions = {
        method: "PUT",
        headers: {
            "Content-Type": "application/json",
        },
        body: JSON.stringify(body),
    };

    const response = await fetch(`/api/user/${id}`, requestOptions)
    const data = await response.json()
    if (!response.ok) {
        throw new Error(data.message);
    }
    toast.success(data.message);

    return data.user;
}

async function deleteUser(user: User) {

    const confirmDelete = window.confirm(`Are you sure you want to delete ${user.name} (${user.email})?`);
    if (!confirmDelete) {
        return false;
    }

    const requestOptions = {
        method: "DELETE",
        headers: {
            "Content-Type": "application/json",
        },
    };

    const response = await fetch(`/api/user/${user.id}`, requestOptions)
    const data = await response.json()
    if (!response.ok) {
        throw new Error(data.message);
    }
    toast.success(data.message);
    return true;
}

const defaultSortFunction = (a: User, b: User) => {
    if (a.role !== "admin" && b.role === "admin") {
        return 1;
    } else if (a.role === "admin" && b.role !== "admin") {
        return -1;
    } else if (a.role === "user" && b.role !== "user") {
        return 1;
    } else if (b.role === "user" && a.role !== "user") {
        return -1;
    } else if (!a.permissions.active && b.permissions.active) {
        return 1;
    } else if (a.permissions.active && !b.permissions.active) {
        return -1;
    }
    return a.name.localeCompare(b.name)
}

type UserManagementConfig = {
    shownColumns: { [key: string]: boolean }
}

function UserManagementComponent() {
    const [users, setUsers] = useState<User[]>([]);
    const [nrens, setNrens] = useState<Nren[]>([]);
    const { user: loggedInUser, setUser: updateCurrentUser } = useContext(userContext);
    const [sortColumn, setSortColumn] = useState({ column: "ID", asc: true });
    const [filter, setFilter] = useState("");

    const { setConfig, user_management } = useConfig<UserManagementConfig>("user_management");

    const setColumnVisibility = (column: string, show: boolean) => {
        const newConfig = user_management ?? {};
        const columnConfig = newConfig?.shownColumns;

        if (!columnConfig) {
            setConfig({ ...newConfig, shownColumns: { [column]: show } });
            return;
        }
        setConfig({ ...newConfig, shownColumns: { ...columnConfig, [column]: show } });
    }

    const getColumnVisibility = (column: string): boolean => {
        const config = user_management;
        if (!config) {
            return true;
        }
        const columnConfig = config["shownColumns"];
        if (!columnConfig) {
            return true;
        }
        const columnShown = columnConfig[column];
        return columnShown ?? true;
    }

    useEffect(() => {
        fetchUsers().then((userList) => {
            setUsers(userList);
        });

        fetchNrens().then((nrenList) => {
            setNrens(nrenList.sort((a, b) => a.name.localeCompare(b.name)))
        })
    }, []);

    const handleInputChange = (event: React.ChangeEvent<HTMLInputElement | HTMLSelectElement>, user: User) => {

        const index = users.findIndex((u) => u.id === user.id);
        const updatedUsers = [...users];

        const { name } = event.target;

        const update = {};

        if (name === "active") {
            update[name] = (event.target as HTMLInputElement).checked
        } else {
            update[name] = event.target.value;
        }

        updateUser(user.id, update).then((user) => {
            if (user.id === loggedInUser.id) {
                updateCurrentUser(user);
            } else {
                updatedUsers[index] = user;
                setUsers(updatedUsers);
            }
        }).catch((error) => {
            toast.error(error.message);
        });
    };

    const findNrenId = (value) => {
        return nrens.find((nren) => {
            return nren.id == value || nren.name == value
        })?.id
    }

    const sortByActive = (a: User, b: User) => {
        if (a.permissions.active && !b.permissions.active) {
            return 1;
        } else if (!a.permissions.active && b.permissions.active) {
            return -1;
        } else {
            return defaultSortFunction(a, b)
        }
    }

    const sortBy = (property: keyof User) => {
        return (a: User, b: User) => {
            const valueA = a[property];
            const valueB = b[property];

            if (property === "nrens") {
                if (a.nrens.length === 0 && b.nrens.length === 0) {
                    return defaultSortFunction(a, b)
                } else if (a.nrens.length === 0) {
                    return -1;
                } else if (b.nrens.length === 0) {
                    return 1;
                } else {
                    return a.nrens[0].localeCompare(b.nrens[0])
                }
            }

            if (typeof valueA !== "string" || typeof valueB !== "string") {
                return defaultSortFunction(a, b)
            }
            const sortVal = valueA.localeCompare(valueB);

            if (sortVal === 0) {
                return defaultSortFunction(a, b)
            }
            return sortVal;
        }
    }

    const columns = ["ID", "Active", "Role", "Email", "Full Name", "OIDC Sub", "NREN", "Actions"]
    const sortFunctions = {
        [columns[1]]: sortByActive,
        [columns[2]]: sortBy("role"),
        [columns[3]]: sortBy("email"),
        [columns[4]]: sortBy("name"),
        [columns[6]]: sortBy("nrens"),
    }

    const setSort = (column) => {
        if (column === sortColumn.column) {
            setSortColumn({ column, asc: !sortColumn.asc })
        }
        else {
            setSortColumn({ column, asc: true })
        }
    }

    // This controls which column to put a sort icon on.
    // We only add it to columns that don"t use the default sort function, or to the ID column.
    const ariaSort = {}
    if (Array.from(Object.keys(sortFunctions)).includes(sortColumn.column)) {
        ariaSort[sortColumn.column] = ({ "aria-sort": sortColumn.asc ? "ascending" : "descending" })
    } else {
        ariaSort[columns[0]] = ({ "aria-sort": sortColumn.asc ? "ascending" : "descending" })
    }

    const sortFunction = sortFunctions[sortColumn.column] ?? defaultSortFunction

    const filteredUsers = filter ? users.filter((u) => u.email.includes(filter) || u.name.includes(filter)) : users
    const userList = filteredUsers.filter((u) => u.id !== loggedInUser.id).sort(sortFunction)
    if (!sortColumn.asc) {
        userList.reverse()
    }
    return (
        <>
            <SurveySidebar />
            <Toaster />
            <Container className="py-5 grey-container">
                <Row className="d-flex justify-content-center align-items-center flex-column">
                    <div className="text-center w-100 mb-3">
                        <h3>User Management Page</h3>
                    </div>
                    <Accordion className="mb-3" style={{ width: "30rem" }}>
                        <Accordion.Item eventKey="0">
                            <Accordion.Header>
                                <span className="me-2">Column Visibility</span>
                                <small className="text-muted">Choose which columns to display</small>
                            </Accordion.Header>
                            <Accordion.Body>
                                <Form.Control as="div" className="p-3">
                                    <small className="text-muted mb-2 d-block">
                                        Select which columns you want to display in the table below. Unchecked columns will be hidden.
                                    </small>
                                    <div className="d-grid" style={{
                                        gridTemplateColumns: "repeat(auto-fill, minmax(150px, 1fr))",
                                        gap: "10px"
                                    }}>
                                        {columns.map((column) => (
                                            <Form.Check
                                                key={column}
                                                type="checkbox"
                                                id={`column-${column}`}
                                                label={column}
                                                checked={getColumnVisibility(column)}
                                                onChange={(e) => setColumnVisibility(column, e.target.checked)}
                                            />
                                        ))}
                                    </div>
                                </Form.Control>
                            </Accordion.Body>
                        </Accordion.Item>
                    </Accordion>
                    <InputGroup className="mb-3" style={{ width: "30rem" }}>
                        <InputGroup.Text id="search-text">
                            Search
                        </InputGroup.Text>
                        <Form.Control
                            placeholder="Search by email/name"
                            aria-label="Search"
                            onInput={debounce((e) => setFilter((e.target as HTMLInputElement).value), 200)}
                        />
                        <Button variant="outline-secondary" onClick={() => {
                            setFilter("")
                        }}>Clear</Button>
                    </InputGroup>
                </Row>

                <div className="d-flex justify-content-center">
                    <div style={{ maxWidth: "100rem" }}>
                        <Table className="user-management-table" bordered >
                            <colgroup>
                                {getColumnVisibility(columns[0]) && <col span={1} style={{ width: "8rem" }} />}
                                {getColumnVisibility(columns[1]) && <col span={1} style={{ width: "3rem" }} />}
                                {getColumnVisibility(columns[2]) && <col span={1} style={{ width: "4.5rem" }} />}
                                {getColumnVisibility(columns[3]) && <col span={1} style={{ width: "7rem" }} />}
                                {getColumnVisibility(columns[4]) && <col span={1} style={{ width: "5rem" }} />}
                                {getColumnVisibility(columns[5]) && <col span={1} style={{ width: "5rem" }} />}
                                {getColumnVisibility(columns[6]) && <col span={1} style={{ width: "6rem" }} />}
                                {getColumnVisibility(columns[7]) && <col span={1} style={{ width: "3rem" }} />}
                            </colgroup>
                            <thead>
                                <tr>
                                    {columns.map((column) => (
                                        getColumnVisibility(column) &&
                                        <th key={column} {...ariaSort[column]} onClick={() => setSort(column)} className="sortable fixed-column" style={{ border: "1px solid #ddd" }}>
                                            {column}
                                        </th>
                                    ))}
                                </tr>
                            </thead>
                            <tbody>
                                {(!filter ? [loggedInUser] : []).concat(userList).map(user => (
                                    <tr key={user.id} style={{ fontWeight: user.id == loggedInUser.id ? "bold" : "normal" }}>
                                        {getColumnVisibility(columns[0]) && <td style={{ border: "1px dotted #ddd" }}>{user.id}</td>}
                                        {getColumnVisibility(columns[1]) && <td style={{ border: "1px dotted #ddd" }}>
                                            {user.id == loggedInUser.id ? <FaCheck /> : <input
                                                type="checkbox"
                                                name="active"
                                                checked={user.permissions.active}
                                                onChange={(event) => handleInputChange(event, user)}
                                            />}
                                        </td>}
                                        {getColumnVisibility(columns[2]) && <td style={{ border: "1px dotted #ddd" }}>
                                            {user.id == loggedInUser.id ? (user.role.charAt(0).toUpperCase()
                                                + user.role.slice(1)) : <select
                                                    name="role"
                                                    defaultValue={user.role}
                                                    onChange={(event) => handleInputChange(event, user)}
                                                    style={{ width: "100%" }}>
                                                <option value="admin">Admin</option>
                                                <option value="user">User</option>
                                                <option value="observer">Observer</option>
                                            </select>}
                                        </td>}

                                        {getColumnVisibility(columns[3]) && <td style={{ border: "1px dotted #ddd" }}>
                                            {user.email}
                                        </td>}
                                        {getColumnVisibility(columns[4]) && <td style={{ border: "1px dotted #ddd" }}>
                                            {user.name}
                                        </td>}
                                        {getColumnVisibility(columns[5]) && <td style={{ border: "1px dotted #ddd" }}>
                                            {user.oidc_sub}
                                        </td>}
                                        {getColumnVisibility(columns[6]) && <td style={{ border: "1px dotted #ddd" }}>
                                            <select
                                                name="nren"
                                                multiple={false}
                                                value={user.nrens.length > 0 ? findNrenId(user.nrens[0]) : ""}
                                                onChange={(event) => handleInputChange(event, user)}>
                                                <option value="">
                                                    Select NREN
                                                </option>
                                                {nrens.map((nren) => (
                                                    <option key={nren.id} value={nren.id}>
                                                        {nren.name}
                                                    </option>
                                                ))}
                                            </select>
                                        </td>}
                                        {getColumnVisibility(columns[7]) && <td style={{ border: "1px dotted #ddd" }}>
                                            {user.id !== loggedInUser.id && <Button variant="danger" onClick={async () => {
                                                if (user.id === loggedInUser.id) {
                                                    toast.error("You cannot delete yourself")
                                                    return
                                                }
                                                const success = await deleteUser(user)
                                                if (success) {
                                                    setUsers(users.filter((u) => u.id !== user.id))
                                                }
                                            }}>
                                                Delete</Button>}
                                        </td>}
                                    </tr>
                                ))}
                            </tbody>
                        </Table>
                    </div>
                </div>
            </Container>
        </>
    );

}

export default UserManagementComponent;
