# annotations import is required for sqlalchemy annotations to work properly
from __future__ import annotations

import logging
import sqlalchemy as sa

from datetime import datetime
from enum import Enum
from typing import Dict, Any, List, Optional
from typing_extensions import Annotated
from uuid import UUID

from sqlalchemy.orm import Mapped, mapped_column, relationship
from sqlalchemy.schema import ForeignKey
from sqlalchemy.types import JSON


from compendium_v2.db import db
from compendium_v2.db.auth_model import User
from compendium_v2.db.presentation_models import NREN


logger = logging.getLogger(__name__)


int_pk = Annotated[int, mapped_column(primary_key=True)]
int_pk_fkNREN = Annotated[int, mapped_column(ForeignKey("nren.id"), primary_key=True)]
int_pk_fkSurvey = Annotated[int, mapped_column(ForeignKey("survey.year"), primary_key=True)]
uuid_nullable_fkUser = Annotated[Optional[UUID], mapped_column(ForeignKey("user.id"))]
json = Annotated[Dict[str, Any], mapped_column(JSON)]

survey_response_pk_fkSurvey = Annotated[int, mapped_column(ForeignKey("survey_response.survey_year"), primary_key=True)]
survey_response_pk_fkNREN = Annotated[int, mapped_column(ForeignKey("survey_response.nren_id"), primary_key=True)]


# Unfortunately flask-sqlalchemy doesnt fully support DeclarativeBase yet.
# See https://github.com/pallets-eco/flask-sqlalchemy/issues/1140
# mypy: disable-error-code="name-defined"

class SurveyStatus(Enum):
    closed = "closed"
    open = "open"
    preview = "preview"
    published = "published"


class ResponseStatus(Enum):
    started = "started"
    completed = "completed"


RESPONSE_NOT_STARTED = "not started"
RESPONSE_NOT_COMPLETED = "did not respond"


class Survey(db.Model):
    __tablename__ = 'survey'
    year: Mapped[int_pk]
    survey: Mapped[json]
    responses: Mapped[List[SurveyResponse]] = relationship(back_populates="survey")
    status: Mapped[SurveyStatus]


class SurveyResponse(db.Model):
    __tablename__ = 'survey_response'
    nren_id: Mapped[int_pk_fkNREN]
    nren: Mapped[NREN] = relationship(lazy='joined')
    survey_year: Mapped[int_pk_fkSurvey]
    survey: Mapped[Survey] = relationship(lazy='joined', back_populates="responses")
    answers: Mapped[json]
    status: Mapped[ResponseStatus]
    locked_by: Mapped[uuid_nullable_fkUser]
    locked_by_user: Mapped[User] = relationship(lazy='joined')
    lock_uuid: Mapped[Optional[UUID]]
    valid: Mapped[bool] = mapped_column(default=False)
    notes: Mapped[SurveyNotes] = relationship("SurveyNotes", back_populates="survey", uselist=False, lazy='joined')

    @property
    def lock_description(self):
        if self.locked_by_user:
            return f"Locked by: {self.locked_by_user.fullname} ({self.locked_by_user.email})"
        return ""

    @property
    def lock_username(self):
        if self.locked_by_user:
            return self.locked_by_user.fullname
        return ""


class SurveyNotes(db.Model):
    __tablename__ = 'survey_notes'

    __table_args__ = (
        sa.ForeignKeyConstraint(['survey_year', 'nren_id'], ['survey_response.survey_year', 'survey_response.nren_id']),
    )

    survey_year: Mapped[int] = mapped_column(primary_key=True)
    nren_id: Mapped[int] = mapped_column(primary_key=True)

    notes: Mapped[str] = mapped_column(nullable=False)
    created_at: Mapped[datetime] = mapped_column(default=datetime.now, nullable=False)
    updated_at: Mapped[datetime] = mapped_column(default=datetime.now, onupdate=datetime.now, nullable=False)
    survey = relationship("SurveyResponse", back_populates="notes", lazy='joined')
