"""
survey_publisher_legacy_excel
=============================

This module loads the survey data from before 2022 from a legacy Excel file.
Missing info is filled in from the survey db for some questions.
Returns the data ready to be used to generate new-style surveys.

"""
from __future__ import annotations
import itertools
from sqlalchemy import select
from collections import defaultdict

from compendium_v2.db import db
from compendium_v2.publishers import helpers, excel_parser
from compendium_v2.survey_db import model as survey_model


def budget(nren_dict):
    nren_by_id = {nren.id: nren for nren in nren_dict.values()}
    data = db.session.scalars(select(survey_model.Nrens))
    inserts = defaultdict(dict)
    for nren in data:
        for budget in nren.budgets:
            abbrev = nren.abbreviation.upper()
            year = budget.year

            if abbrev not in nren_dict:
                continue

            budget_entry = {
                'nren': nren_dict[helpers.map_nren(abbrev)],
                'nren_id': nren_dict[helpers.map_nren(abbrev)].id,
                'budget': float(budget.budget),
                'year': year
            }
            inserts[nren_dict[helpers.map_nren(abbrev)].id][year] = budget_entry

    # Import the data from excel sheet to database
    exceldata = excel_parser.fetch_budget_excel_data()

    for abbrev, budget, year in exceldata:
        if abbrev not in nren_dict:
            continue

        budget_entry = {
            'nren': nren_dict[helpers.map_nren(abbrev)],
            'nren_id': nren_dict[helpers.map_nren(abbrev)].id,
            'budget': budget,
            'year': year
        }
        inserts[nren_dict[helpers.map_nren(abbrev)].id][year] = budget_entry

    for nren_id, year_data in inserts.items():
        for year, budget_entry in year_data.items():
            nren = nren_by_id[nren_id]
            yield ('budget', nren, nren.id, year, budget_entry['budget'])


def funding(nren_dict):
    data = excel_parser.fetch_funding_excel_data()
    for (abbrev, year, client_institution,
            european_funding,
            gov_public_bodies,
            commercial, other) in data:

        if abbrev not in nren_dict:
            continue

        nren = nren_dict[helpers.map_nren(abbrev)]

        _data = {
            'client_institutions': round(client_institution, 2),
            'european_funding': round(european_funding, 2),
            'commercial': round(commercial, 2),
            'other': round(other, 2),
            'gov_public_bodies': round(gov_public_bodies, 2),
        }

        if nren.name.lower() == 'heanet':
            if year == 2016:
                _data['other'] = 1.0
            if year == 2018:
                _data['other'] = 3.0

        if sum(_data.values()) == 0:
            continue

        yield ('income_sources', nren, nren.id, year, _data)


def charging_structure(nren_dict):
    data = excel_parser.fetch_charging_structure_excel_data()

    for (abbrev, year, charging_structure) in data:
        if abbrev not in nren_dict:
            continue

        nren = nren_dict[helpers.map_nren(abbrev)]

        if charging_structure:
            charging_structure = charging_structure.name

        yield ('charging_mechanism', nren, nren.id, year, charging_structure)


def staffing(nren_dict):
    staff_data = list(excel_parser.fetch_staffing_excel_data())

    nren_staff_map = {}
    for (abbrev, year, permanent_fte, subcontracted_fte) in staff_data:
        if abbrev not in nren_dict:
            continue

        nren = nren_dict[helpers.map_nren(abbrev)]
        nren_staff_map[(nren.id, year)] = {
            'nren': nren,
            'nren_id': nren.id,
            'year': year,
            'permanent_fte': permanent_fte,
            'subcontracted_fte': subcontracted_fte,
            'technical_fte': 0,
            'non_technical_fte': 0
        }

    function_data = excel_parser.fetch_staff_function_excel_data()
    for (abbrev, year, technical_fte, non_technical_fte) in function_data:
        if abbrev not in nren_dict:
            continue

        nren = nren_dict[helpers.map_nren(abbrev)]
        if (nren.id, year) in nren_staff_map:
            nren_staff_map[(nren.id, year)]['technical_fte'] = technical_fte
            nren_staff_map[(nren.id, year)]['non_technical_fte'] = non_technical_fte
        else:
            nren_staff_map[(nren.id, year)] = {
                'nren': nren,
                'nren_id': nren.id,
                'year': year,
                'permanent_fte': 0,
                'subcontracted_fte': 0,
                'technical_fte': technical_fte,
                'non_technical_fte': non_technical_fte
            }

    for nren_staff_model in nren_staff_map.values():

        nren = nren_staff_model['nren']
        year = nren_staff_model['year']
        staff_type_data = {
            'permanent_fte': nren_staff_model['permanent_fte'],
            'subcontracted_fte': nren_staff_model['subcontracted_fte'],
        }

        staff_roles_data = {
            'technical_fte': nren_staff_model['technical_fte'],
            'nontechnical_fte': nren_staff_model['non_technical_fte'],
        }

        yield ('staff_employment_type', nren, nren.id, year, staff_type_data)

        yield ('staff_roles', nren, nren.id, year, staff_roles_data)


def ecprojects(nren_dict):
    ecproject_data = excel_parser.fetch_ecproject_excel_data()

    by_nren_year = defaultdict(lambda: defaultdict(set))
    for (abbrev, year, project) in ecproject_data:
        if abbrev not in nren_dict:
            continue

        nren = nren_dict[helpers.map_nren(abbrev)]

        by_nren_year[nren][year].add(project)

    for nren, year_projects in by_nren_year.items():
        for year, projects in year_projects.items():
            projects = [{'ec_project_name': project} for project in projects]
            yield ('ec_project_names', nren, nren.id, year, projects)


def parent_organizations(nren_dict):
    organization_data = excel_parser.fetch_organization_excel_data()
    for (abbrev, year, org) in organization_data:
        if abbrev not in nren_dict:
            continue

        nren = nren_dict[helpers.map_nren(abbrev)]
        yield ('parent_organization', nren, nren.id, year, 'Yes')
        yield ('parent_organization_name', nren, nren.id, year, org)


def traffic_volume(nren_dict):
    traffic_data = excel_parser.fetch_traffic_excel_data()
    for (abbrev, year, from_external, to_external, from_customers, to_customers) in traffic_data:
        if abbrev not in nren_dict:
            continue

        nren = nren_dict[helpers.map_nren(abbrev)]
        if nren.name == 'CESNET':
            # COMP-447: correct CESNET traffic data for 2019
            if year == 2019:
                to_customers = 222766

        yield ('traffic_estimate', nren, nren.id, year, {
            'from_customers': from_customers,
            'from_external': from_external,
            'to_customers': to_customers,
            'to_external': to_external
        })


def connected_proportion(nren_dict):
    remit = excel_parser.fetch_remit_excel_data()
    nr_connected = excel_parser.fetch_nr_connected_excel_data()
    market_share = excel_parser.fetch_market_share_excel_data()
    users_served = excel_parser.fetch_users_served_excel_data()

    data_by_nren = defaultdict(lambda: defaultdict(dict))

    for key in itertools.chain(remit.keys(), nr_connected.keys(), market_share.keys(), users_served.keys()):
        (abbrev, year, user_category) = key
        if abbrev not in nren_dict:
            continue

        nren = nren_dict[helpers.map_nren(abbrev)]

        covered = remit.get(key)
        if covered:
            covered = covered.name
        market_share_percentage = market_share.get(key)
        _nr_connected = nr_connected.get(key)
        nr_of_users = users_served.get(key)

        result = {}
        if covered:
            result['covered'] = covered
        if market_share_percentage:
            result['market_share_percentage'] = market_share_percentage
        if _nr_connected:
            result['nr_connected'] = _nr_connected
        if nr_of_users:
            result['nr_of_users'] = nr_of_users

        existing = data_by_nren[nren][year].get(user_category)
        if existing:
            existing.update(result)
        else:
            data_by_nren[nren][year][user_category] = result

    for nren, data in data_by_nren.items():
        for year, category_data in data.items():
            connectivity_proportions = {}
            for user_category, user_data in category_data.items():
                connectivity_proportions[user_category.name] = user_data

            yield ('connectivity_proportions', nren, nren.id, year, connectivity_proportions)


def connectivity_level(nren_dict):
    typical_speeds = excel_parser.fetch_typical_speed_excel_data()
    highest_speeds = excel_parser.fetch_highest_speed_excel_data()
    highest_speed_proportions = excel_parser.fetch_highest_speed_proportion_excel_data()
    data_by_nren = defaultdict(lambda: defaultdict(dict))

    for key in itertools.chain(typical_speeds.keys(), highest_speeds.keys(), highest_speed_proportions.keys()):
        (abbrev, year, user_category) = key
        if abbrev not in nren_dict:
            continue

        nren = nren_dict[helpers.map_nren(abbrev)]

        typical_speed = typical_speeds.get(key)
        highest_speed = highest_speeds.get(key)
        highest_speed_connection_percentage = highest_speed_proportions.get(key)

        result = {}
        if typical_speed:
            result['typical_speed'] = typical_speed
        if highest_speed:
            result['highest_speed'] = highest_speed
        if highest_speed_connection_percentage:
            result['highest_speed_connection_percentage'] = highest_speed_connection_percentage

        existing = data_by_nren[nren][year].get(user_category)
        if existing:
            existing.update(result)
        else:
            data_by_nren[nren][year][user_category] = result

    for nren, data in data_by_nren.items():
        for year, category_data in data.items():
            connectivity_level = {}
            for user_category, user_data in category_data.items():
                connectivity_level[user_category.name] = user_data
            yield ('connectivity_level', nren, nren.id, year, connectivity_level)


def connection_carrier(nren_dict):
    carriers = excel_parser.fetch_carriers_excel_data()
    data_by_nren = defaultdict(lambda: defaultdict(dict))

    for key, carry_mechanism in carriers.items():
        (abbrev, year, user_category) = key
        if abbrev not in nren_dict:
            continue

        nren = nren_dict[helpers.map_nren(abbrev)]

        result = {}
        if carry_mechanism:
            result['carry_mechanism'] = carry_mechanism.name

        existing = data_by_nren[nren][year].get(user_category)
        if existing:
            existing.update(result)
        else:
            data_by_nren[nren][year][user_category] = result

    for nren, data in data_by_nren.items():
        for year, category_data in data.items():
            traffic_carriers = {}
            for user_category, user_data in category_data.items():
                traffic_carriers[user_category.name] = user_data
            yield ('traffic_carriers', nren, nren.id, year, traffic_carriers)


def connectivity_growth(nren_dict):
    growth = excel_parser.fetch_growth_excel_data()
    data_by_nren = defaultdict(lambda: defaultdict(dict))
    for key, growth_percent in growth.items():
        (abbrev, year, user_category) = key
        if abbrev not in nren_dict:
            continue
        nren = nren_dict[helpers.map_nren(abbrev)]

        result = {}
        if growth_percent:
            result['growth_rate'] = growth_percent

        existing = data_by_nren[nren][year].get(user_category)
        if existing:
            existing.update(result)
        else:
            data_by_nren[nren][year][user_category] = result

    for nren, data in data_by_nren.items():
        for year, category_data in data.items():
            traffic_growth = {}
            for user_category, user_data in category_data.items():
                traffic_growth[user_category.name] = user_data
            yield ('traffic_growth', nren, nren.id, year, traffic_growth)


def connectivity_load(nren_dict):
    averages = excel_parser.fetch_average_traffic_excel_data()
    peaks = excel_parser.fetch_peak_traffic_excel_data()

    all_entry_keys = set()
    all_entry_keys.update(averages.keys())
    all_entry_keys.update(peaks.keys())
    data_by_nren = defaultdict(lambda: defaultdict(dict))
    for key in all_entry_keys:
        (abbrev, year, user_category) = key
        if abbrev not in nren_dict:
            continue

        nren = nren_dict[helpers.map_nren(abbrev)]

        result = {}
        average = averages.get(key, (None, None))
        peak = peaks.get(key, (None, None))
        average_from_institutions_to_network = average[0]
        average_to_institutions_from_network = average[1]
        peak_from_institutions_to_network = peak[0]
        peak_to_institutions_from_network = peak[1]

        if average_from_institutions_to_network is not None:
            result['average_from_institutions_to_network'] = average_from_institutions_to_network
        if average_to_institutions_from_network is not None:
            result['average_to_institutions_from_network'] = average_to_institutions_from_network
        if peak_from_institutions_to_network is not None:
            result['peak_from_institutions_to_network'] = peak_from_institutions_to_network
        if peak_to_institutions_from_network is not None:
            result['peak_to_institutions_from_network'] = peak_to_institutions_from_network

        existing = data_by_nren[nren][year].get(user_category)
        if existing:
            existing.update(result)
        else:
            data_by_nren[nren][year][user_category] = result

    for nren, data in data_by_nren.items():
        for year, category_data in data.items():
            traffic_load = {}
            for user_category, user_data in category_data.items():
                traffic_load[user_category.name] = user_data
            yield ('traffic_load', nren, nren.id, year, traffic_load)


def remote_campuses(nren_dict):
    campuses = excel_parser.fetch_remote_campuses_excel_data()

    for (abbrev, year, connectivity, country, connected_to_r_e) in campuses:
        if abbrev not in nren_dict:
            continue

        nren = nren_dict[helpers.map_nren(abbrev)]

        yield ('remote_campuses', nren, nren.id, year, "Yes" if connectivity else "No")
        if country:
            yield ('remote_campuses_specifics', nren, nren.id, year, [
                {'country': country, 'connected': 'Yes' if connected_to_r_e else 'No'}
            ])


def dark_fibre_lease(nren_dict):
    data_rows = excel_parser.fetch_dark_fibre_iru_excel_data()
    iru_durations = excel_parser.fetch_iru_duration_excel_data()

    for (abbrev, year, iru, length_in_country, length_out_country) in data_rows:
        if abbrev not in nren_dict:
            continue

        nren = nren_dict[helpers.map_nren(abbrev)]

        dark_fibre_lease = iru
        dark_fibre_lease_duration = iru_durations.get((helpers.map_nren(abbrev), year))

        yield ('dark_fibre_lease', nren, nren.id, year, "Yes" if dark_fibre_lease else "No")

        if dark_fibre_lease_duration is not None:
            yield ('dark_fibre_lease_duration', nren, nren.id, year, dark_fibre_lease_duration)
        if length_in_country is not None:
            yield ('dark_fibre_lease_kilometers_inside_country', nren, nren.id, year, length_in_country)
        if length_out_country is not None:
            yield ('dark_fibre_lease_kilometers_outside_country', nren, nren.id, year, length_out_country)


def dark_fibre_installed(nren_dict):
    data_rows = excel_parser.fetch_dark_fibre_installed_excel_data()
    for (abbrev, year, installed, length) in data_rows:
        if abbrev not in nren_dict:
            continue

        nren = nren_dict[helpers.map_nren(abbrev)]

        yield ('dark_fibre_nren', nren, nren.id, year, "Yes" if installed else "No")
        yield ('dark_fibre_nren_kilometers_inside_country', nren, nren.id, year, length)


def passive_monitoring(nren_dict):
    data_rows = excel_parser.fetch_passive_monitoring_excel_data()
    for (abbrev, year, monitoring, method) in data_rows:
        if abbrev not in nren_dict:
            continue

        nren = nren_dict[helpers.map_nren(abbrev)]

        if monitoring or method:
            yield ('passive_monitoring', nren, nren.id, year, "Yes")
        else:
            yield ('passive_monitoring', nren, nren.id, year, "No")
        if method:
            yield ('passive_monitoring_tech', nren, nren.id, year, method.name)


def capacity(nren_dict):
    largest_data_rows = excel_parser.fetch_largest_link_capacity_excel_data()
    typical_data_rows = excel_parser.fetch_typical_backbone_capacity_excel_data()

    by_nren = defaultdict(dict)

    for key in itertools.chain(largest_data_rows.keys(), typical_data_rows.keys()):
        (abbrev, year) = key
        if abbrev not in nren_dict:
            continue

        to_add = (abbrev, year)
        by_nren[to_add].update({
            'max_capacity': largest_data_rows.get(key, by_nren[to_add].get('max_capacity')),
            'typical_capacity': typical_data_rows.get(key, by_nren[to_add].get('typical_capacity'))
        })

    for (abbrev, year), data in by_nren.items():
        max_capacity = data.get('max_capacity')
        typical_capacity = data.get('typical_capacity')

        nren = nren_dict[helpers.map_nren(abbrev)]

        if max_capacity:
            yield ('max_capacity', nren, nren.id, year, max_capacity)
        if typical_capacity:
            yield ('typical_capacity', nren, nren.id, year, typical_capacity)


def non_r_e_peers(nren_dict):
    data_rows = excel_parser.fetch_non_r_e_peers_excel_data()
    for (abbrev, year, nr_of_non_r_and_e_peers) in data_rows:
        if abbrev not in nren_dict:
            continue

        nren = nren_dict[helpers.map_nren(abbrev)]

        if nr_of_non_r_and_e_peers:
            yield ('non_r_and_e_peers', nren, nren.id, year, nr_of_non_r_and_e_peers)


def ops_automation(nren_dict):
    data_rows = excel_parser.fetch_ops_automation_excel_data()
    for (abbrev, year, automation, specifics) in data_rows:
        if abbrev not in nren_dict:
            continue

        nren = nren_dict[helpers.map_nren(abbrev)]

        if automation:
            name = automation.name[0].upper() + automation.name[1:]
            yield ('operational_process_automation', nren, nren.id, year, name)

        if specifics:
            yield ('operational_process_automation_tools', nren, nren.id, year, specifics)


def fetch_data():
    # requires being in a flask app context when called
    nren_dict = helpers.get_uppercase_nren_dict()
    yield from budget(nren_dict)
    yield from funding(nren_dict)
    yield from charging_structure(nren_dict)
    yield from staffing(nren_dict)
    yield from ecprojects(nren_dict)
    yield from parent_organizations(nren_dict)
    yield from traffic_volume(nren_dict)

    yield from connected_proportion(nren_dict)
    yield from connectivity_level(nren_dict)
    yield from connection_carrier(nren_dict)
    yield from connectivity_growth(nren_dict)
    yield from connectivity_load(nren_dict)
    yield from remote_campuses(nren_dict)

    yield from dark_fibre_lease(nren_dict)
    yield from dark_fibre_installed(nren_dict)
    yield from passive_monitoring(nren_dict)
    yield from capacity(nren_dict)
    yield from non_r_e_peers(nren_dict)
    yield from ops_automation(nren_dict)
