import json
import jsonschema
import pytest
import pathlib
from sqlalchemy import select
import compendium_v2.routes.schemas
from compendium_v2.routes.api import models
from compendium_v2.db import db
from compendium_v2.db.presentation_models import PreviewYear, NREN, NRENService, Service

TEST_DATA_DIR = pathlib.Path(__file__).parent / 'data' / 'presentation_models_data'


def _get_data(model):
    filename = TEST_DATA_DIR / f'{model.__name__}.json'
    if not filename.exists():
        raise FileNotFoundError(
            f'Data file for {model.__name__} not found at {filename}. Run test_data_dump.py to generate it.')
    with open(filename, 'r') as f:
        return json.load(f)


def setup(app, model):
    with app.app_context():
        # set up the presentation model data for NREN first, as the instances are needed for other models
        nren_data = _get_data(NREN)
        nrens = {
            nren['id']: NREN(**nren)
            for nren in nren_data
        }
        db.session.add_all(nrens.values())
        db.session.commit()

        if model == NREN:
            return

        if model == NRENService:
            services = {s.name_key: s for s in db.session.scalars(select(Service)).all()}

        def _create_instance(model, data):
            nren = nrens[data['nren_id']]
            if model == NRENService:
                service = services[data['service_key']]
                return model(nren=nren, service=service, **data)

            return model(nren=nren, **data)

        data = _get_data(model)
        instances = [_create_instance(model, instance) for instance in data]
        db.session.add_all(instances)
        db.session.commit()


@pytest.mark.parametrize('model, url, schema', models)
def test_presentation_model_routes(app, mocked_admin_user, client, model, url, schema, nren_services):
    nren_services(app)

    setup(app, model)
    rv = client.get(
        f'/api{url}',
        headers={'Accept': ['application/json']})
    assert rv.status_code == 200
    result = json.loads(rv.data.decode('utf-8'))
    jsonschema.validate(result, getattr(compendium_v2.routes.schemas, schema))
    assert result

    if model == NREN:
        # as long as we have NREN data, we're good
        return

    years_of_data = set([datapoint['year'] for datapoint in result])
    assert len(years_of_data) > 0

    with app.app_context():
        # set preview year to the first year of data, to test the preview functionality
        db.session.add(PreviewYear(year=sorted(years_of_data)[0]))
        db.session.commit()

    # with ?preview, the response should be the same as the first request
    rv = client.get(
        f'/api{url}?preview',
        headers={'Accept': ['application/json']})
    assert rv.status_code == 200
    result = json.loads(rv.data.decode('utf-8'))
    jsonschema.validate(result, getattr(compendium_v2.routes.schemas, schema))
    years_of_data_preview = set([datapoint['year'] for datapoint in result])
    assert years_of_data_preview == years_of_data

    # without ?preview, the response should be missing the preview year
    rv = client.get(
        f'/api{url}',
        headers={'Accept': ['application/json']})
    assert rv.status_code == 200
    result = json.loads(rv.data.decode('utf-8'))
    jsonschema.validate(result, getattr(compendium_v2.routes.schemas, schema))
    if len(years_of_data) > 1:
        assert result
    else:
        assert not result
    years_of_data_without_preview = set([datapoint['year'] for datapoint in result])
    assert len(years_of_data_preview - years_of_data_without_preview) == 1
